#!/usr/bin/perl

# TODO: useful statistics at end of encoding
# TODO: profiling of metadata to facilitate various smart and dumb media players

use strict;
use warnings;

use open ':utf8';

use Encode;
use File::Find qw(find);
use File::Path qw(mkpath);
use File::Spec::Functions qw(catpath rel2abs splitpath);
use Getopt::ArgvFile home=>1;
use Getopt::Long qw(GetOptions);
use Parallel::ForkManager;
use Sys::CPU;

# name of this program and associated version
my $progname = 'flac2vorbis';
my $version = '0.5';

# command line options (defaulted)
my $encopts = '--quality=5';
my $metaflac = 'metaflac';
my $oggenc = 'oggenc';
my $processes = Sys::CPU::cpu_count();

# command line options (undefined)
my $quiet; # false
my $overwrite; # false
my $verbose; # false

# command line arguments
my $flac_dir;
my $vorbis_dir;

# return true if value defined and not empty
sub is_value {
  return (defined($_[0]) && length($_[0]) > 0);
}

my $header; # false

sub header {

  # do nothing if already output or user asked to be quiet
  return if (defined($header) || defined($quiet));

  print STDERR "\n";
  print STDERR "$progname $version: Transcodes FLAC files to Ogg Vorbis format en masse.\n";
  print STDERR "Copyright 2008-2009 Paul C. Bryan.\n";
  print STDERR "\n";

  $header = 1;
}

# output informative message if not quiet
sub msg {
  header; print STDERR "$_[0]\n" if (!defined($quiet));
}

# display error message if not quiet, then exit with error code
sub error {
  undef $quiet; msg("$_[0]\n"); exit 1;
}

# output diagnostic if verbosity enabled
sub diag {
  msg($_[0]) if (defined($verbose));
}

# display system call as diagnostic, then execute system call
sub sys {
  diag($_[0]); system($_[0]) == 0 or error("Error calling $_[0].");
}

# trims spaces from start and end of string
sub trim {
  return undef if (!defined($_[0]));
  my $v = $_[0];
  $v =~ s/^[[:space:]]+//; $v =~ s/[[:space:]]+$//;
  return $v;
}

# escape special characters to make appropriate for quoted program arguments
sub escq {
  return undef if (!defined($_[0]));
  my $v = $_[0];
  $v =~ s/\"/\\\"/g;
  $v =~ s/\$/\\\$/g;
  return '"' . $v . '"';
}

sub trail_slash {
  return undef if (!defined($_[0]));
  my $v = $_[0];
  $v =~ s/([^\/])$/$1\//;
 return $_[0];
}

# returns "no" if boolean false (undefined)
sub noboolean {
  return (!defined($_[0]) ? "no" : "");
}

# extract year from date
sub extract_year {
  return undef if (!defined($_[0]));
  my $v = $_[0];
  if ($v =~ m/(\d{4})/) { return $v; }
  return undef;
}

# display usage and exit
sub usage {

  undef $quiet;
  header();

	print STDERR <<EOF;
Usage: $progname [options] flac_dir [vorbis_dir]

OPTIONS:
  --encopts=<opts>   Options to pass to encoder [$encopts].
  --help             Display this help message.
  --oggenc=<prog>    Program to encode to Ogg Vorbis [$oggenc].
  --metaflac=<prog>  Program to use to manage metadata in FLAC files [$metaflac].
  --[no]overwrite    Overwrite previous extract of tracks [${\(noboolean($overwrite))}overwrite].
  --processes=<num>  Number of concurrent encoder processes [$processes].
  --quiet            Quiet operation, no screen output.
  --verbose          Verbose output of progress information.

CONFIGURATION FILE:
  You can store persistent command-line options in a file called .$progname in
  your home directory. They will be read as defaults prior to processing any
  passed command-line options.

EOF

	exit 1;
}

# parse command line options (may override defaults or initialization)
sub get_options {

  GetOptions(
    'encopts=s'   => \$encopts,
    'help'        => \&usage,
    'metaflac=s'  => \$metaflac,
    'nooverwrite' => sub { undef $overwrite; },
    'oggenc=s'    => \$oggenc,
    'overwrite'   => sub { $overwrite = 1; },
    'processes=i' => \$processes,
    'quiet'       => sub { $quiet = 1; undef $verbose; },
    'verbose'     => sub { $verbose = 1; undef $quiet; },
  ) or usage;

  # first argument mandatory, second optional
  usage() if (@ARGV < 1 || @ARGV > 2);

  # first parameter: directory to find flac files (make absolute)
  $flac_dir = rel2abs($ARGV[0]);

  # second (optional, defaults to flac directory)
  $vorbis_dir = (@ARGV >= 2 ? rel2abs($ARGV[1]) : $flac_dir);

  # ensure director(ies) specified are indeed directories
  error("$flac_dir is not a directory.") if (!-d $flac_dir);
  error("$vorbis_dir is not a writeable directory.") if !(-d $vorbis_dir && -w $vorbis_dir);

  # directories should have trailing slashes
  $flac_dir = trail_slash($flac_dir);
  $vorbis_dir = trail_slash($vorbis_dir);
}

# get metadata from flac file
sub get_metadata {

  my $flac = $_[0];

  my %metadata = ();

  open(METAFLAC, "$metaflac --no-utf8-convert --export-tags-to=- " . escq($flac) . '|');

  while (<METAFLAC>) {
  
    chomp;

  utf8::encode($_);

# FIXME: multiple tags of same name overwritten...
    # assign metadata value to appropriate hash entry
    if (m/^(.*)=(.*)/) {
     $metadata{$1} = $2 if (is_value($1) && is_value($2));
    }
  }

  close(METAFLAC);

  return %metadata;
}

sub comment {
  return defined($_[1]) ? ' --comment=' . escq("$_[0]=$_[1]") : '';
}

sub process_file {

  # get the complete pathname to the file
  my $flac_file = $_[0];

  # get flac file path relative to specified flac directory
  my $relative_flac = substr($flac_file, length($flac_dir));

  # create vorbis file path from flac relative file path
  my $vorbis_path = $relative_flac;
  $vorbis_path =~ s/(.flac$)/.ogg/;
  $vorbis_path = $vorbis_dir . $vorbis_path;

  # do not overwrite unless explicitly requested
  if (-e $vorbis_path && !defined $overwrite) {
    diag("Ignoring $vorbis_path (not overwriting existing file).");
    return;
  }

  # a most useful message
  msg("Writing $vorbis_path.");

  # get metadata from the file
  my %metadata = get_metadata($flac_file);

  my $title = $metadata{'TITLE'};

  my $featuring = $metadata{'FEATURING'};
  $title .= " (feat. $featuring)" if (is_value($title) && is_value($featuring));

  my $version = $metadata{'VERSION'};
  $title .= " ($version)" if (is_value($title) && is_value($version));

  # get base directory for vorbis file
  my ($vol, $dir, $file) = splitpath($vorbis_path);
  my $vorbis_base = catpath($vol, $dir, '');

  # make base directory if it does not exist
  eval { mkpath($vorbis_base) };
  if ($@) { error("Couldn't create $vorbis_base directory for files ($@)."); }

  my $cmd = 'cat ' . escq($flac_file) . " | $oggenc --discard-comments" .
   ((!defined $verbose) ? ' --quiet' : '') .
   ((defined $encopts) ? " $encopts" : '') .
   comment('ARTIST', $metadata{'ARTIST'}) .
   comment('ALBUMARTIST', $metadata{'ALBUMARTIST'}) .
   comment('DATE', $metadata{'DATE'}) .
   comment('ALBUM', $metadata{'ALBUM'}) .
   comment('TRACKNUMBER', $metadata{'TRACKNUMBER'}) .
   comment('TITLE', $title) .
   comment('TRACKID', $metadata{'TRACKID'}) .
   comment('MUSICBRAINZ_TRACKID', $metadata{'MUSICBRAINZ_TRACKID'}) .
   ' --output=' . escq($vorbis_path) . ' -';

  sys($cmd);
}

# place to store found files
my @files = ();

sub push_file {

  # get the complete pathname to the file
  my $file = $_;

  # if file is a directory, ignore it
  return if (-d $file);

  # if file does not have flac extension, ignore it
  return if (!($file =~ m/\.flac$/));

  # file must be in flac directory because we're eventually going to compute relative name
  error("$file not in $flac_dir") if (index($file, $flac_dir) != 0);

  push(@files, $file);
}

# iterate through files in flac directory, calling process_file for each
sub process_files {

  # get all files up-front
  find({wanted =>\&push_file, no_chdir=> 1, follow=>1, follow_skip=>2, dangling_symlinks=>0}, $flac_dir);

  # some deterministic way of iterating through the files found
  @files = sort { $a cmp $b } @files;

  my $pm = new Parallel::ForkManager($processes);

  for my $file (@files) {
    $pm->start() and next;
    process_file($file);
    $pm->finish();
  }

  $pm->wait_all_children();
}

# output program header
header();

# parse command line options
get_options();

# iterate through all files in the flac directory, processing each
process_files();

msg "\nDone.";
